#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2017 Canonical Ltd.
#
# This file is part of the Apt Stresstest Charm for Juju.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import argparse
import datetime
import glob
import os
import socket
import subprocess
import sys
from yaml import safe_load


class Timer(object):
    def __init__(self):
        self._timer = None
        self._delta = None

    def start(self):
        self._timer = datetime.datetime.now()

    def stop(self):
        self._delta = datetime.datetime.now() - self._timer

    @property
    def time(self):
        return self._delta.total_seconds()


def devnull():
    return open(os.devnull, 'w')


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument('-v', '--verbose', help="increase output verbosity.",
                        action="store_true")

    parser.add_argument('-p', "--package", help="package to download from repositories.",
                        default="bash")

    parser.add_argument('-f', "--hosts-file", help="local path for chrooted /etc/hosts configuration file.",
                        default=None)

    parser.add_argument('-s', '--send-to', help="where to write the output, e.g. 127.0.0.1:8094",
                        default=None)

    parser.add_argument('--update-timeout', default=20, type=int,
                        help="apt-get update timeout in seconds. If exceeded metric value will be set to 0",
                        )

    parser.add_argument('--download-timeout', default=120, type=int,
                        help="apt-get download timeout in seconds. If exceeded metric value will be set to 0",
                        )
    parser.add_argument('--series', type=int,
                        help="Default Ubuntu release/series to test",
                        )

    return parser.parse_args()


def apt_update(ip, chroot=None, timeout=30):
    cmd = ["apt-get", "-qq", "-c", get_apt_config_path(ip), "update"]
    try:
        duration = run(cmd, timeout=timeout)
    except subprocess.TimeoutExpired:
        duration = 0
    return duration


def apt_download(package, ip, no_cache=False, timeout=120):
    cmd = ["apt-get", "-qq", "-c", get_apt_config_path(ip)]
    if no_cache:
        cmd.extend(["-o", "acquire::http::no-cache=true"])
    cmd.extend(["download", "--", package])
    try:
        duration = run(cmd, timeout=timeout)
    except subprocess.TimeoutExpired:
        duration = 0
    apt_cleanup()
    return duration


def apt_cleanup():
    cmd = ["apt-get", "-qq", "clean"]
    run(cmd)
    for file in glob.glob('/tmp/*.deb'):
        os.remove(file)


def read_mirror_config_file(config_path):
    with open(config_path, 'r') as f:
        data = safe_load(f.read())
        # There may be a cleaner way to split this data apart but we don't want
        # to mess around with the config format and this is workable for now.
        # For reference the config format looks like so:
        # - {us-central1.gce.archive.ubuntu.com: 35.184.213.5}
        # - {us-central1.gce.archive.ubuntu.com: 35.202.116.96}
        # - {us-central1.gce.archive.ubuntu.com: 35.192.35.27}
        # - {us-central1.gce.archive.ubuntu.com: 35.202.204.188}
        domain = list(data[0].keys())[0]
        output = []
        for i in data:
            output.append(list(i.values())[0])
        return domain, output


def run(cmd, stderr=None, timeout=120):
    # The output from run doesn't appear to be used anywhere, though it wouldn't
    # be a bad idea to print it for debugging, so returning the time spent seems
    # relatively safe.
    timer = Timer()
    timer.start()
    p = subprocess.Popen(cmd, stdout=devnull(), stderr=stderr, cwd="/tmp")
    try:
        output, _ = p.communicate(timeout=timeout)
    except subprocess.TimeoutExpired:
        # Give process time to release file handles and propagate exception for handling
        p.wait()
        raise
    timer.stop()
    return timer.time


def render_influx(test, domain, ip, value, metric_name="apt_transaction_duration_seconds"):
    line_fmt = "{metric_name},test={test},mirror={domain},ip={ip} value={value}"
    return line_fmt.format(
        metric_name=metric_name,
        test=test,
        domain=domain,
        ip=ip,
        value=value
    )


def send_to_influx(send_to, metrics):
    (host, port) = send_to.split(':')
    port = int(port)
    s = socket.socket()
    s.connect((host, port))
    with s.makefile(mode='w') as dest:
        dest.write(metrics)


def output_results(result, verbose=False, send_to=None):
    if verbose:
        print(result)
    if send_to:
        if verbose:
            print("DEBUG: Sending data to influxdb: {}".format(send_to))
        send_to_influx(send_to, result)


def run_tests(domain, ip, packages, args):

    output_results(render_influx("update_1", domain, ip,
                   apt_update(ip, timeout=args.update_timeout)),
                   verbose=args.verbose, send_to=args.send_to)
    output_results(render_influx("update_2", domain, ip,
                   apt_update(ip, timeout=args.update_timeout)),
                   verbose=args.verbose, send_to=args.send_to)
    for index, package in enumerate(packages):
        test_name = "cached_download_{}_{}".format(1 + index, package)
        output_results(render_influx(test_name, domain, ip,
                       apt_download(package, ip, timeout=args.download_timeout)),
                       verbose=args.verbose, send_to=args.send_to)
        test_name = "uncached_download_{}_{}".format(1 + index, package)
        output_results(render_influx(test_name, domain, ip,
                       apt_download(package, ip, no_cache=True, timeout=args.download_timeout)),
                       verbose=args.verbose, send_to=args.send_to)


def get_apt_config_path(ip):
    return "/etc/apt-stresstest/{!s}-apt.conf".format(ip)


# Taken from charm-helpers
# https://github.com/juju/charm-helpers/blob/master/charmhelpers/core/host_factory/ubuntu.py#L69
def lsb_release():
    """Return /etc/lsb-release in a dict"""
    d = {}
    with open('/etc/lsb-release', 'r') as lsb:
        for l in lsb:
            k, v = l.split('=')
            d[k.strip()] = v.strip()
    return d


def write_file(dest_path, content):
    dest = ''
    if os.path.exists(dest_path):
        with open(dest_path, 'r') as f:
            dest = f.read()

    if content == dest:
        return False

    with open(dest_path, 'w') as f:
        f.write(content)


def write_apt_config(domain, ip, series=None):
    if not series:
        series = lsb_release()['DISTRIB_CODENAME']
    apt_config_path = get_apt_config_path(ip)
    sources_list_path = "/etc/apt-stresstest/{!s}-sources.list".format(ip)
    # Create the config dir if it doesn't exist
    if not os.path.isdir("/etc/apt-stresstest/"):
        os.mkdir("/etc/apt-stresstest/", 0o755)

    # Write per unit mirror configs/sources.list if they don't exist or differs
    apt_config = 'Dir::Etc::SourceList "/etc/apt-stresstest/{!s}-sources.list";\n'.format(ip)
    write_file(apt_config_path, apt_config)
    sources_list = "deb http://{!s}/ubuntu/ {!s} main universe\n".format(ip, series)
    sources_list += "deb http://{!s}/ubuntu/ {!s}-updates main universe\n".format(ip, series)
    write_file(sources_list_path, sources_list)


def main():
    timer = Timer()
    timer.start()
    args = parse_args()

    if args.hosts_file:
        domain, ips = read_mirror_config_file(args.hosts_file)
    else:
        ips = []

    for ip in ips:
        write_apt_config(domain, ip, args.series)
        run_tests(
            domain=domain,
            ip=ip,
            packages=args.package.split(','),
            args=args,
        )
    timer.stop()
    output_results(render_influx("N/A", "N/A", "N/A", timer.time, "apt_transaction_total_duration_seconds"),
                   verbose=args.verbose, send_to=args.send_to)
    output_results(render_influx("N/A", "N/A", "N/A", len(ips), "apt_mirror_units_count"),
                   verbose=args.verbose, send_to=args.send_to)


if __name__ == '__main__':
    sys.exit(main())
