# -*- coding: utf-8 -*-
# Copyright 2017 Canonical Ltd.
#
# This file is part of the Apt Stresstest Charm for Juju.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import shutil
from charmhelpers.core import hookenv, host
from charms.reactive import hook, when, when_not, set_state, remove_state
from yaml import safe_load, dump


APT_STRESSTEST_CRONTAB_TEMPLATE = """\
#
# This file is managed by Juju.  Attempt no changes here.
#

SHELL=/bin/sh
PATH=/usr/local/sbin:/usr/local/bin:/sbin:/bin:/usr/sbin:/usr/bin
* * * * * root /usr/bin/run-one timeout {stresstest_timeout} /usr/local/bin/test_apt_mirrors.py {args} -s 127.0.0.1:8094
"""

MIRROR_CONFIG_PATH = "/etc/apt-stresstest-mirrors.yaml"

IMAGES_CACHE_PATH = "/var/cache/apt-stresstest/images"


def stress_test_args():
    config = hookenv.config()
    args = []

    mirror_config = config.get('stresstest-mirror-config')

    if mirror_config:
        args.extend(['-f', MIRROR_CONFIG_PATH])

    if config.get('apt-package'):
        packages = ','.join(config.get('apt-package').split())
        args.extend(['-p', packages])

    if config.get('update-timeout'):
        args.extend(['--update-timeout', config.get('update-timeout')])

    if config.get('download-timeout'):
        args.extend(['--download-timeout', config.get('download-timeout')])

    return ' '.join(args)


def install_script():
    charm_dir = hookenv.charm_dir()

    if os.path.exists('/usr/local/bin/test_apt_mirrors.py'):
        os.remove('/usr/local/bin/test_apt_mirrors.py')

    shutil.copy(
        os.path.join(charm_dir, 'files/test_apt_mirrors.py'),
        '/usr/local/bin/test_apt_mirrors.py'
    )


def install_crontab():
    config = hookenv.config()
    host.write_file(
        '/etc/cron.d/apt-stresstest',
        content=APT_STRESSTEST_CRONTAB_TEMPLATE.format(
            args=stress_test_args(),
            stresstest_timeout=config.get('stresstest-timeout'),
        ).encode('utf-8')
    )


@when_not('apt-stresstest.host_file')
def provision_stresstest_mirror_config():
    config = hookenv.config()
    host_file_config = config.get('stresstest-mirror-config')
    if host_file_config:
        with open(MIRROR_CONFIG_PATH, 'w') as w:
            dump(safe_load(host_file_config), w)

    set_state('apt-stresstest.host_file')


@hook('upgrade-charm', 'config-changed')
def config_changed():
    reconfigure()


@when('config.changed')
def reconfigure():
    remove_state('apt-stresstest.host_file')
    remove_state('apt-stresstest.installed')


@when(
    'apt-stresstest.host_file',
    'apt.installed.python3-requests',
    'apt.installed.run-one',
    'apt.installed.schroot',
    'apt.installed.simplestreams',
    'apt.installed.ubuntu-cloudimage-keyring'
)
@when_not('exec.available')
def status_wait_for_telegraf():
    hookenv.status_set('blocked', 'Waiting for telegraf relation.')


@when(
    'apt-stresstest.host_file',
    'apt.installed.python3-requests',
    'apt.installed.run-one',
    'apt.installed.schroot',
    'apt.installed.simplestreams',
    'apt.installed.ubuntu-cloudimage-keyring',
    'exec.available'
)
@when_not('apt-stresstest.installed')
def install_apt_stresstest(exec_relation):
    install_script()
    install_crontab()
    os.makedirs(IMAGES_CACHE_PATH, exist_ok=True)

    set_state('apt-stresstest.installed')
    hookenv.status_set('active', 'apt-stresstest installed.')
